/*
 * Various utility routines put togheter by Tito <tito-wolit@tiscali.it> .
 *
 * Copyright (C) 2002-2003-2004 by Tito Ragusa <tito-wolit@tiscali.it>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */

/* This is to avoid a compiler warning for `index` with -Wshadow and Gtk-+-1.2 libs */
#ifndef USE_GTK_2_0
	#ifdef IS_LINUX
		#undef  __USE_BSD
		#include <string.h>
		#define  __USE_BSD
	#endif
#else
	#include <string.h>
#endif

/********************* Debug utilities *****************************************/

/* Various error message routines.
 *
 * inspired by bb_verror_msg(), bb_perror_msg() and bb_error_msg()
 * from busybox's libbb.
 * Copyright (C) 1999-2003 by Erik Andersen <andersen@codepoet.org>
 *
 * These functions understand %m as strerror(errno).
 *
 */

#include <stdio.h>
#include <stdarg.h>
#include <stdlib.h>

#ifdef L_err_msg
void err_msg(const char *fmt, ...)
{
	va_list p;
	fflush(stdout);
	#ifdef PROGNAME
	fprintf(stderr, "%s: ", PROGNAME);
	#endif
	va_start(p, fmt);
	vfprintf(stderr,fmt, p);
	va_end(p);
	putc('\n', stderr);
}
#endif

#ifdef L_msg
void msg(const char *fmt, ...)
{
	va_list p;

	fflush(stdout);
	#ifdef PROGNAME
	fprintf(stdout, "%s: ", PROGNAME);
	#endif
	va_start(p, fmt);
	vfprintf(stdout,fmt, p);
	va_end(p);
	putc('\n', stdout);
}
#endif

/****************End of Debug utilities ****************************************/

/*
 * Utility routines.
 *
 * Copyright (C) 1999,2000,2001 by Erik Andersen <andersee@debian.org>
 * Copyright 1996, 1997 Linux International.
 * Contributed by Richard Henderson <rth@tamu.edu>
 */

#include <sys/types.h>
#include <sys/stat.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <fcntl.h>

/* xalloc_die (C) 2003 Tito Ragusa <tito-wolit@tiscali.it> GPL */

#ifdef L_xalloc_die
void xalloc_die (void * ptr)
{
	if (ptr == NULL)
	{
		/* Report a failure to allocate memory and terminate the program.
		This function does not use fprintf to avoid further memory allocation attempts.
		This idea was found in  opag, an option parser generator Copyright (C) 2002, 2004 Martin Dickopp */
		write (STDERR_FILENO, "memory exhausted\n", sizeof("memory exhausted\n") );
		exit(EXIT_FAILURE);
	}
}
#endif

#ifdef L_xmalloc
void *xmalloc(size_t size)
{
	void *ptr = malloc((size) ? size : 1);
	xalloc_die(ptr);
	return ptr;
}
#endif

#ifdef L_xrealloc
void *xrealloc(void *ptr, size_t size)
{
	/* It avoids to free ptr if size==0 */
	/* if ptr==NULL it does malloc(size) */
	ptr = realloc(ptr, (size) ? size : 1);
	xalloc_die(ptr);
	return ptr;
}
#endif

#ifdef L_xcalloc
void *xcalloc(size_t nmemb, size_t size)
{
	/* It avoids nmemb==0 and size==0 */
	void *ptr = calloc(nmemb ? (nmemb) : 1, (size) ? size : 1);
	xalloc_die(ptr);
	return ptr;
}
#endif

#ifdef L_xfree
void *xfree(void *ptr)
{
	/*	avoids freeing a NULL pointer and sets the pointer to NULL
		after it has been freed.*/
	if(ptr)
	{
		free(ptr);
		ptr=NULL;
	}
	#ifdef DEBUG
	else
		err_msg("Attempt to free NULL pointer");
	#endif
	return ptr;
}
#endif

/* xopen_die (C) 2003 Tito Ragusa <tito-wolit@tiscali.it> GPL */

#ifdef L_xopen_die
void xopen_die(const char *path)
{
	#ifdef DEBUG		
	err_msg("failed to open: %s: %m", path);
	#endif
	exit(EXIT_FAILURE);
}
#endif

#ifdef L_xfopen
FILE *xfopen(const char *path, const char *mode)
{
	FILE *fp;
	if ((fp = fopen(path, mode)) == NULL)
		xopen_die(path);
	return fp;
}
#endif

#ifdef L_xpopen
FILE *xpopen(const char *path, const char *mode)
{
	FILE *fp;
	if ((fp = popen(path, mode)) == NULL)
		xopen_die(path);
	return fp;
}
#endif

/*
 * xgetcwd.c -- return current directory with unlimited length
 * Copyright (C) 1992, 1996 Free Software Foundation, Inc.
 * Written by David MacKenzie <djm@gnu.ai.mit.edu>.
 *
 * Special function for busybox written by Vladimir Oleynik <dzo@simtreas.ru>
 */

#include <stdlib.h>
#include <errno.h>
#include <unistd.h>
#include <limits.h>
#include <sys/param.h>

/* Amount to increase buffer size by in each try. */
#define PATH_INCR 32

/* Return the current directory, newly allocated, arbitrarily long.
   Return NULL and set errno on error.
   If argument is not NULL (previous usage allocate memory), call free()
*/

#ifdef L_xgetcwd
char * xgetcwd (char *cwd)
{
	char *ret;
	unsigned path_max;
	
	path_max = (unsigned) PATH_MAX;
	path_max += 2;                /* The getcwd docs say to do this. */
	
	if(cwd==0)
		cwd = xmalloc (path_max);
	
	while ((ret = getcwd (cwd, path_max)) == NULL && errno == ERANGE)
	{
		path_max += PATH_INCR;
		cwd = xrealloc (cwd, path_max);
	}
	
	if (ret == NULL)
	{
		cwd=xfree(cwd);
		#ifdef DEBUG     
		err_msg("getcwd()");
		#endif      
		return NULL;
	}
	
	return cwd;
}
#endif

/*
* Copyright (C) 2003 by Tito Ragusa <tito-wolit@tiscali.it>
*
* This function checks strings for illegal chars that could
* be misinterpetrated if passed to a shell by a system() call.
*/

#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <ctype.h>

#ifdef L_illegal_chars_in_string
char *illegal_chars_in_string(char *string,int substitute)
{
	int illegal_chars=0;
	int i=0;


	/* skip white space */
	while (string && isspace(*string)) {
		++string;
	}

	if(string==NULL)
		return NULL;

	/* Don't allow leading dash */
	if(string[0]== '-')
	{
		if(substitute==1)
			string[0]='_';
		illegal_chars++;
	}

	for(i=0 ; i<=(int)xstrlen(string) ; i++)
	{
		switch(string[i])
		{
			case ' '	:
			/*case '/'	: legal in paths */ 
			/*case 92	: legal in urls  */ /*case '\':*/
			case '*'	:
			case '?'	:
			case ':'	:
			case ';'	:
			case ','	:
			case 39		: /*case ''': */
			case '`'	:
			case '<'	:
			case '>'	:
			case '"'	:
			case '!'	:
			case '$'	:
			case '%'	:
			case '&'	:
			case '='	:
			case '~'	:
			case '^'	:
			case '|'	:
			case '@'	:
			case '#'	:
			case '('	:
			case ')'	:
			case '['	:
			case ']'	:
			case '{'	:
			case '}'	:
				illegal_chars++;
				if(substitute==1)
					string[i]='_';
				break;
			case '-'	:
				/* Don't allow double dash */
			case '.'	:
				/* Don't allow double dot */
				if(string[i]==string[i+1])
				{
					illegal_chars++;
					if(substitute==1)
					{
						string[i]='_';
						string[i+1]='_';
					}
				}
				break;
			/*esac*/
		}/*switch*/
	}/*for*/
	if(illegal_chars>0 && substitute==0)
		return NULL;
	return string;
}
#endif

/*
* Copyright (C) 2003 by Tito Ragusa <tito-wolit@tiscali.it>
*
*  This function reads chars from a fd until EOF storing them into
*  a properly sized buffer and returns a pointer to this buffer .
*  The buffer should be freed .
*/

#ifdef L_fgetc_buf
char * fgetc_buf(FILE *fd)
{
	int c, i=0;
	int size=1024;
	char *buffer=NULL;

	buffer=xmalloc(size);
	while((c=fgetc(fd))!=EOF)
	{
		buffer[i]=c;
		if(i==(size - 2))
			buffer=xrealloc(buffer,size+=1024);
		i++;
	}
	buffer[i]='\0';
	return buffer;
}
#endif
/*
* Copyright (C) 2004 by Tito Ragusa <tito-wolit@tiscali.it>
*
*  This function reads chars from a fd until EOF storing them into
*  a buffer of given size and returns a pointer to this buffer or NULL on EOF .
*  The buffer should be freed .
*/

#ifdef L_xfgetc_nbuf
char * xfgetc_nbuf(FILE *fd, int size)
{
	int c, i=0;
	char *buffer;

	buffer=xmalloc(size);

	while((c=fgetc(fd))!=EOF)
	{
		buffer[i]=c;
		i++;
		if(i==(size - 2))
			break;
	}
	
	/* NULL terminate the buffer */
	buffer[i]='\0';
	
	/*	Read to the end, and on the next iteration return NULL
		so we know the job is done */
	if(i==0 && c==EOF)
	{
		buffer=xfree(buffer);
		/*return NULL;*/
	}
	return buffer;
}
#endif

/*
*   Copyright (C) 2003 by Tito Ragusa <tito-wolit@tiscali.it>
*
*	Function for formatted output conversion,  that disallows
*	string truncation.
*/

#include <stdio.h>
#include <stdlib.h>

#ifdef L_xsnprintf
int xsnprintf(char *dest, size_t size, const char *fmt, ...)
{
	int ret;
	va_list p;

	/*The functions snprintf and vsnprintf do not write more  than
	size  bytes (including the trailing '\0').  If the output was truncated
	due to this limit then the return value is  the  number  of  characters
	(not  including the trailing '\0') which would have been written to the
	final string if enough space had been available. Thus, a  return  value
	of  size  or  more means that the output was truncated.
	If an output error is encountered, a negative  value  is returned.
	The  glibc  implementation of the functions snprintf and vsnprintf con-
	forms to the C99 standard, i.e.,  behaves  as  described  above,  since
	glibc version 2.1. Until glibc 2.0.6 they would return -1 when the out-
	put was truncated.
	Linux  libc5 adds a new  conversion character m, which outputs strerror(errno).
	*/

	fflush(stdout);
	va_start(p, fmt);
	ret=vsnprintf(dest,size,fmt, p);
	va_end(p);

	if(ret < 0)
	{
		#ifdef DEBUG		
		err_msg("xsnprintf(): failed");
		#endif
		exit(EXIT_FAILURE);
	}
	#if (__GLIBC__ >= 2) && (__GLIBC_MINOR__ >= 1)
	if ((unsigned int)ret >= size)
	{
		#ifdef DEBUG
		err_msg("xsnprintf(): output truncated");
		#endif		
		exit(EXIT_FAILURE);
	}
	#endif
	return(ret);
}
#endif

/* Misc utility functions.
*   Copyright 2000 Keith Owens <kaos@ocs.com.au>
*
*   This file is part of the Linux modutils.
*/

#include <stdlib.h>

#ifdef L_xstrncat
char *xstrncat(char *dest, const char *src, size_t size)
{
	unsigned int ldest = xstrlen(dest);
	unsigned int lsrc = xstrlen(src);
	if ((size - ldest - 1) < lsrc)
	{
		#ifdef DEBUG		
		err_msg("xstrncat(): destination overflow");
		#endif
		exit(EXIT_FAILURE);
	}
	memcpy(dest+ldest, src, lsrc+1);
	return(dest);
}
#endif

/*  Misc utility functions.
*   Copyright 1996, 1997 Linux International.
*   Contributed by Richard Henderson <rth@tamu.edu>
*
*   This file is part of the Linux modutils.
*   Hacked by Tito 
*/

#include <stdio.h>
#include <stdlib.h>

#ifdef L_xstrdup
char *xstrdup(const char *s)
{
	char *n = strdup(s);
	xalloc_die(n);
	return n;
}
#endif

/*
* Copyright (C) 2003 by Tito Ragusa <tito-wolit@tiscali.it>
*
* This function changes the current directory to that specified in path
* and exits on failure or returns -1 accordingly to int flag.
*/

#include <unistd.h>
#include <stdio.h>
#include <errno.h>
/*#include <string.h>*/

#ifdef L_xchdir
int xchdir(char *path, int flag)
{
	if((chdir(path))==0)
		return EXIT_SUCCESS;
	#ifdef DEBUG	
	err_msg("cannot chdir() to %s: %m",path);
	#endif
	if(flag==EXIT_ON_FAILURE)
		exit(EXIT_FAILURE);
	return EXIT_ERROR;
}
#endif

/*
* Copyright (C) 2003 by Tito Ragusa <tito-wolit@tiscali.it>
*
* This function  attempts to create a directory named pathname.
* mode  specifies the permissions to use. It is modified by the process's
* umask in the usual way and exits on failure or returns -1 accordingly to int flag.
*/

#include <unistd.h>
#include <stdio.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <errno.h>

#ifdef L_xmkdir
int xmkdir(const char *pathname , mode_t mode, int flag)
{
	if((mkdir(pathname, mode))==0)
		return EXIT_SUCCESS;
	#ifdef DEBUG	
	err_msg("cannot mkdir(): %s with mode %#o: %m",pathname,mode);
	#endif	
	if(flag==EXIT_ON_FAILURE)
		exit(EXIT_FAILURE);
	return EXIT_ERROR;
}
#endif

/*
* Copyright (C) 2003 by Tito Ragusa <tito-wolit@tiscali.it>
*
* The  owner of the file specified by path is changed.This function
* exits on failure or returns -1 accordingly to int flag.
*/

#include <unistd.h>
#include <stdio.h>
#include <sys/types.h>
#include <errno.h>

#ifdef L_xchown
int xchown(const char *path, uid_t owner, gid_t group, int flag)
{
	if((chown(path, owner, group ))==0)
		return EXIT_SUCCESS;
	#ifdef DEBUG		
	err_msg("cannot chown(): %s to uid %d and gid %d: %m\n",path,owner,group);
	#endif	
	if(flag==EXIT_ON_FAILURE)
		exit(EXIT_FAILURE);
	return EXIT_ERROR;
}
#endif

/*
* Copyright (C) 2003 by Tito Ragusa <tito-wolit@tiscali.it>
*
*  The  mode of the file given by path is changed.This function
*  exits on failure or returns -1 accordingly to int flag.
*/

#include <unistd.h>
#include <stdio.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <errno.h>

#ifdef L_xchmod
int xchmod(const char *path, mode_t mode, int flag)
{
	if((chmod(path, mode ))==0)
		return EXIT_SUCCESS;
	#ifdef DEBUG		
	err_msg("cannot chmod(): %s : %m\n",path);
	#endif	
	if(flag==EXIT_ON_FAILURE)
		exit(EXIT_FAILURE);
	return EXIT_ERROR;
}
#endif

/*
* Copyright (C) 2003 by Tito Ragusa <tito-wolit@tiscali.it>
*
*   The  xfclose  function  dissociates the named stream from its underlying
*   file or set of functions. If stream is NULL or stdin this function returns -1.
*/

#include <stdio.h>

#ifdef L_xfclose_nostdin
int xfclose_nostdin(FILE *fd)
{
	if(fd && fd != stdin)
		return fclose(fd);
	#ifdef DEBUG
	if(fd==NULL)
		err_msg("cannot close: fd == NULL");
	if(fd==stdin)
		err_msg("cannot close: fd == stdin");
	#endif
	return EXIT_ERROR;
}
#endif

/*
* Copyright (C) 2003 by Tito Ragusa <tito-wolit@tiscali.it>
*
*    The  xpclose_nostdin  function waits for the associated process to terminate and
*    returns the exit status of the command as returned by wait4.
*    If stream is NULL or stdin this function returns -1.
*/

#include <stdio.h>

#ifdef L_xpclose_nostdin
int xpclose_nostdin(FILE *fd)
{
	if(fd && fd != stdin)
		return pclose(fd);
	#ifdef DEBUG
	if(fd==NULL)
		err_msg("cannot close: fd == NULL");
	if(fd==stdin)
		err_msg("cannot close: fd == stdin");
	#endif
	return EXIT_ERROR;
}
#endif

/*
 * busybox library eXtended function
 *
 * Copyright (C) 2001 Larry Doolittle, <ldoolitt@recycle.lbl.gov>
 *
 * Find out if the last character of a string matches the one given. Don't
 * underrun the buffer if the string length is 0.  Also avoids a possible
 * space-hogging inline of strlen() per usage.
 */


#ifdef L_last_char_is
char * last_char_is(const char *s, int c)
{
	char *sret = (char *)s;
	if (sret) {
		sret = strrchr(sret, c);
		if(sret != NULL && *(sret+1) != 0)
			sret = NULL;
	}
	return sret;
}
#endif

/*
* Copyright (C) 2003 by Tito Ragusa <tito-wolit@tiscali.it>
* Find out if the first character of a string matches the one given. Don't
* underrun the buffer if the string length is 0. Based on last_char_is a
* busybox library eXtended function (C) 2001 Larry Doolittle, <ldoolitt@recycle.lbl.gov>
*/

#ifdef L_first_char_is
char * first_char_is(const char *s, int c)
{
	char *sret = (char *)s;
	if (sret) {
		if(sret !=NULL && sret[0]!=c)
			sret = NULL;
	}
	return sret;
}
#endif

/* Copyright (C) many different people (from busybox libbb).
*
*	If you wrote this, please acknowledge your work.
*	Hacked by Tito.
*
*	Concatenate path and file name to new allocated buffer,
*	not adding '/' if path name already have it.
*/


#ifdef L_xconcat_path_file
char *xconcat_path_file(const char *path,char *filename)
{
	char *buffer;
	char *lc;
	int i=0;
	int size;

	if (!path)
	    path="";
	lc = last_char_is(path, '/');

	if( lc == NULL) i=1;

	while (*filename == '/')
		filename++;
	size=xstrlen(path)+xstrlen(filename)+i+1;
	buffer = xmalloc(size);
	xsnprintf(buffer,size, "%s%s%s", path, (lc==NULL)? "/" : "", filename);

	return buffer;
}
#endif

/*
*   Copyright (C) (C) 2003  Vladimir Oleynik  <dzo@simtreas.ru>
*
*   This function covers a special  case for recursive actions of
*   concat_path_file(path, filename) skiping "." and ".." directory
*   entries.
*/

#ifdef L_xconcat_subpath_file
char *xconcat_subpath_file(const char *path, char *f)
{
	if(f && *f == '.' && (!f[1] || (f[1] == '.' && !f[2])))
		return NULL;
	return xconcat_path_file(path, f);
}
#endif

/*
* Copyright (C) 2003 by Tito Ragusa <tito-wolit@tiscali.it>
*
* xputs() writes the string s and a trailing newline to stdout.
* It returns a non-negative number on success, or EOF on error.
* If string is NULL it exits with an error message.
*/

#include <stdio.h>

#ifdef L_xputs
int xputs(const char * string)
{
	if(string==NULL)
	{
		#ifdef DEBUG			
		err_msg("puts(): null string");
		#endif		
		exit(EXIT_FAILURE);
	}
	return puts(string);
}
#endif

#ifdef L_xmake_message
/*
 * From Linux Manpage  2000-10-16  PRINTF(3) EXAMPLES
 *
 * This function allocates a sufficiently large buffer for the given
 * string and prints into it (code correct for both glibc 2.0 and glibc 2.1).
 * It returns a pointer to the malloced buffer or exits if some error occurs.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
*/

#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>

/*
 * NOTE: This function returns a malloced char* that you will have to free
 * yourself. You have been warned.
 */

char * xmake_message(const char *fmt, ...)
{
	/* Guess we need no more than 100 bytes. */
	int n, size = 100;
	char *p;
	va_list ap;

	p = xmalloc(size);
	while (1)
	{
		/* Try to print in the allocated space. */
		va_start(ap, fmt);
		n = vsnprintf (p, size, fmt, ap);
		va_end(ap);
		/* If that worked, return the string. */
		if (n > -1 && n < size)
			return p;
		/* Else try again with more space. */
		if (n > -1)			/* glibc 2.1 */
			size = n + 1;	/* precisely what is needed */
		else				/* glibc 2.0 */
			size *= 2;		/* twice the old size */
		p =  xrealloc (p, size);
	}
}

/*
	NOTES
	The  glibc  implementation of the functions snprintf and vsnprintf con-
	forms to the C99 standard, i.e.,  behaves  as  described  above,  since
	glibc version 2.1. Until glibc 2.0.6 they would return -1 when the out-
	put was truncated.

	RETURN VALUE

	Upon successful return, these functions return the number of characters
	printed  (not  including  the  trailing  '\0'  used  to  end  output to
	strings).  The functions snprintf and vsnprintf do not write more  than
	size  bytes (including the trailing '\0').  If the output was truncated
	due to this limit then the return value is  the  number  of  characters
	(not  including the trailing '\0') which would have been written to the
	final string if enough space had been available. Thus, a  return  value
	of  size  or  more means that the output was truncated. (See also below
	under NOTES.)  If an output error is encountered, a negative  value  is
	returned.

*/
#endif

#ifdef L_get_uid_or_gid

#include <pwd.h>
#include <grp.h>
#include <sys/types.h>

char * get_uid_or_gid (int flag, long idnum)
{

	struct passwd *pw_ent;
	struct group *grp_ent;
	char *message=NULL;
	switch( flag )
	{
		case UID:
			if ( ( pw_ent  = getpwuid (idnum) ) != NULL )
				message = xmake_message( "%s", pw_ent->pw_name);
			else
				message = xmake_message( "%ld", idnum );
			break;
		case GID:
			if ( ( grp_ent = getgrgid (idnum) ) != NULL )
				message = xmake_message ( "%s", grp_ent->gr_name );
			else
				message = xmake_message( "%ld", idnum );
			break;
	}
	return message;
}
#endif

#ifdef L_file_exists
int file_exists(char *filename, struct stat *buf)
{
	struct stat stat_it;

	/* if we stat a broken softlink here it is not displayed as stat returns -1,
	so we use lstat as workaround for this */
	if((filename==NULL) || ((stat(filename,(buf!=NULL)?buf : &stat_it))==-1 && (lstat(filename, (buf!=NULL)?buf : &stat_it))==-1))
	{
			#ifdef DEBUG
			if(filename==NULL)
				err_msg("%s(): filename is NULL", __FUNCTION__);
			else
				err_msg("%s(): unable to stat or lstat: %s: %s", __FUNCTION__, filename ,strerror(errno));
			#endif
			return (-1);
	}
	#ifdef DEBUG
	err_msg("%s(): %s", __FUNCTION__, filename);
	#endif
	return (0);
}
#endif

/*
* Copyright (C) 2004 by Tito Ragusa <tito-wolit@tiscali.it>
*
* xstrlen() returns the length of a string (even if it is NULL) 
*/
#ifdef L_xstrlen
int xstrlen(const char* s)
{
	if(s)
		return strlen(s);
	return 0;
}
#endif
